/*
 * Copyright (c) 2015, 2016, Yutaka Tsutano
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include <iostream>
#include <vector>
#include <regex>
#include <chrono>

#include <boost/graph/graphviz.hpp>
#include <boost/type_erasure/any.hpp>
#include <boost/type_erasure/any_cast.hpp>
#include <boost/type_erasure/operators.hpp>
#include <boost/range/iterator_range.hpp>

#include <jitana/jitana.hpp>
#include <jitana/analysis/call_graph.hpp>
#include <jitana/analysis/def_use.hpp>
#include <jitana/analysis/points_to.hpp>
#include <jitana/analysis/ide.hpp>
#include <jitana/analysis/cha_call_graph.hpp>

void setup_class_loaders(jitana::virtual_machine& vm);
void write_vm_graphs(jitana::virtual_machine& vm);

void run_analysis()
{
    jitana::virtual_machine vm;

    setup_class_loaders(vm);

    // Set the method handle of the entry point.
    std::vector<jitana::jvm_method_hdl> mh_candidates;
    mh_candidates.push_back({{77, "LTest;"}, "main([Ljava/lang/String;)V"});
    mh_candidates.push_back(
            {{22, "Lcom/instagram/android/login/fragment/LoginFragment;"},
             "onCreate(Landroid/os/Bundle;)V"});
    mh_candidates.push_back(
            {{22, "Lcom/instagram/android/login/fragment/LoginFragment;"},
             "onStart()V"});
    mh_candidates.push_back(
            {{22, "Lcom/instagram/android/login/fragment/LoginFragment;"},
             "onResume()V"});
    mh_candidates.push_back(
            {{22, "Lcom/instagram/android/login/fragment/LoginFragment;"},
             "onPause()V"});
    mh_candidates.push_back(
            {{22, "Lcom/instagram/android/login/fragment/LoginFragment;"},
             "onRestart()V"});
    jitana::jvm_method_hdl mh = mh_candidates[1];

    // Load the method and its dependencies.
    auto mv = vm.find_method(mh, true);
    if (!mv) {
        throw std::runtime_error("failed to find the method");
    }
    vm.load_recursive(*mv);

    // Compute the call graph.
    jitana::add_call_graph_edges(vm);

    // Compute the def-use edges.
    std::for_each(vertices(vm.methods()).first, vertices(vm.methods()).second,
                  [&](const jitana::method_vertex_descriptor& v) {
                      add_def_use_edges(vm.methods()[v].insns);
                  });

    std::cout << "Entry point: " << mh << std::endl;

    // Apply CHA analysis to generate CCG.
    std::cout << "Running CHA..." << std::endl;
    auto ccg_cha = make_cha_call_graph(vm, {*mv});

    // Apply points-to analysis to generate CCG.
    std::cout << "Running Points-to..." << std::endl;
    jitana::pointer_assignment_graph pag;
    jitana::contextual_call_graph ccg_pt;
    jitana::update_points_to_graphs(pag, ccg_pt, vm, *mv, true);

    // Generate labeled exploded super graph based on CHA.
    std::cout << "Making CHA-based LESG..." << std::endl;
    auto lesg_cha = make_labeled_exploded_super_graph(vm, ccg_cha);

    // Generate labeled exploded super graph based on points-to.
    std::cout << "Making labeled exploded super graph..." << std::endl;
    auto lesg_pt = make_labeled_exploded_super_graph(vm, ccg_pt);

    // Print stats.
    {
        std::cout << "# of pag vertices: " << num_vertices(pag) << "\n";
        std::cout << "# of pag edges: " << num_edges(pag) << "\n";
        size_t num_p2s = 0;
        for (const auto& v : boost::make_iterator_range(vertices(pag))) {
            num_p2s += pag[v].points_to_set.size();
        }
        std::cout << "# of p2s: " << num_p2s << "\n";
        std::cout << "# of p2s (per vertex): "
                  << (double(num_p2s) / num_vertices(pag)) << "\n";
    }

    std::cout << "Writing PAG..." << std::endl;
    {
        std::ofstream ofs("output/pag.dot");
        jitana::write_graphviz_pointer_assignment_graph(ofs, pag, &vm);
    }

    std::cout << "Writing Points-to CCG..." << std::endl;
    {
        std::ofstream ofs("output/ccg_pt.dot");
        jitana::write_graphviz_contextual_call_graph(ofs, ccg_pt);
    }

    {
        std::ofstream ofs("output/ccg_cha.dot");
        jitana::write_graphviz_contextual_call_graph(ofs, ccg_cha);
    }

    {
        std::ofstream ofs("output/lesg_pt.dot");
        write_graphviz_labeled_exploded_super_graph(ofs, lesg_pt);
    }

    {
        std::ofstream ofs("output/lesg_cha.dot");
        write_graphviz_labeled_exploded_super_graph(ofs, lesg_cha);
    }

    std::cout << "Writing VM graphs..." << std::endl;
    write_vm_graphs(vm);
}

void setup_class_loaders(jitana::virtual_machine& vm)
{
    {
        const auto& filenames
                = {"../../../dex/framework/core.dex",
                   "../../../dex/framework/framework.dex",
                   "../../../dex/framework/framework2.dex",
                   "../../../dex/framework/ext.dex",
                   "../../../dex/framework/conscrypt.dex",
                   "../../../dex/framework/okhttp.dex",
                   "../../../dex/framework/core-junit.dex",
                   "../../../dex/framework/android.test.runner.dex",
                   "../../../dex/framework/android.policy.dex"};
        jitana::class_loader loader(11, "SystemLoader", begin(filenames),
                                    end(filenames));
        vm.add_loader(loader);
    }

    std::string apk_root = "../../../dex/extracted/";

    vm.add_apk(22, apk_root + "data@app@com.instagram.android-1.apk", 11);

    {
        const auto& filenames = {"../../../dex/app/test.dex"};
        jitana::class_loader loader(33, "Test", begin(filenames),
                                    end(filenames));
        vm.add_loader(loader, 11);
    }

    vm.add_apk(44, apk_root + "data@app@jp.bio100.android.superdepth-1.apk",
               11);

    {
        const auto& filenames = {"../../../dex/small_tests/01/01.dex"};
        jitana::class_loader loader(55, "SmallTest01", begin(filenames),
                                    end(filenames));
        vm.add_loader(loader, 11);
    }

    {
        const auto& filenames = {"../../../dex/small_tests/02/02.dex"};
        jitana::class_loader loader(66, "SmallTest02", begin(filenames),
                                    end(filenames));
        vm.add_loader(loader, 11);
    }

    {
        const auto& filenames = {"../../../dex/small_tests/03/03.dex"};
        jitana::class_loader loader(77, "SmallTest03", begin(filenames),
                                    end(filenames));
        vm.add_loader(loader, 11);
    }
}

void write_vm_graphs(jitana::virtual_machine& vm)
{
    {
        std::ofstream ofs("output/loader_graph.dot");
        write_graphviz_loader_graph(ofs, vm.loaders());
    }

    {
        std::ofstream ofs("output/class_graph.dot");
        write_graphviz_class_graph(ofs, vm.classes());
    }

    {
        std::ofstream ofs("output/method_graph.dot");
        write_graphviz_method_graph(ofs, vm.methods());
    }

    {
        std::ofstream ofs("output/field_graph.dot");
        write_graphviz_field_graph(ofs, vm.fields());
    }

    for (const auto& v : boost::make_iterator_range(vertices(vm.methods()))) {
        const auto& ig = vm.methods()[v].insns;
        if (num_vertices(ig) > 0) {
            std::stringstream ss;
            ss << "output/insn/" << vm.methods()[v].hdl << ".dot";
            std::ofstream ofs(ss.str());
            write_graphviz_insn_graph(ofs, ig, &vm);
        }
    }
}

int main()
{
    try {
        run_analysis();
    }
    catch (const std::exception& e) {
        std::cerr << "Error: " << e.what() << "\n\n";
        std::cerr << "Please make sure that ";
        std::cerr << "all dependencies are installed correctly, and ";
        std::cerr << "the DEX files exist.\n";
        return 1;
    }
}
